from typing import List, Tuple

from .base import Operator
from .tokens import END_TOKEN
from .add_sos_eos import AddSosEosAligner
from ..edit import lst2lst
from ..edit.replacement import ReplacementGroup


class AddEndToken(Operator):

    """
    Add end token (<eos>) to a sequence.

    E.g.
    >>> from uttut.pipeline.ops.add_sos_eos import AddSosEos
    >>> op = AddEndToken()
    >>> output_seq, label_aligner = op.transform(
        ['I', 'have', '10.7', 'dollars', '.'])
    >>> output_labels = label_aligner.transform([1, 2, 3, 4, 5])
    >>> output_seq
    ['I', 'have', '10.7', 'dollars', '.', '<eos>']
    >>> output_labels
    [1, 2, 3, 4, 5, 0]
    >>> label_aligner.inverse_transform(output_labels)
    [1, 2, 3, 4, 5]

    """

    _input_type = list
    _output_type = list

    def __init__(self, end_token: str = END_TOKEN):
        self.end_token = self._validate_end_token(end_token)

    def _validate_end_token(self, end_token: str):
        if end_token is None:
            raise ValueError("The end token can not be None.")

        if not isinstance(end_token, str):
            raise TypeError("The end token should be a string.")

        if len(end_token) == 0:
            raise ValueError("The end token should not be an empty string.")

        return end_token

    def _transform(self, input_sequence: List[str]) -> Tuple[List[str], 'LabelAligner']:
        forward_replacement_group = self._gen_forward_replacement_group(input_sequence)
        output_sequence = lst2lst.apply(input_sequence, forward_replacement_group)

        label_aligner = AddSosEosAligner(
            input_sequence=input_sequence,
            edit=forward_replacement_group,
            output_length=len(output_sequence),
        )
        return output_sequence, label_aligner

    def _gen_forward_replacement_group(self, input_lst: List[str]) -> ReplacementGroup:
        seqlen = len(input_lst)
        replacement_group = ReplacementGroup.add_all(
            [(seqlen, seqlen, [self.end_token], 'add-end-token')],
        )
        return replacement_group



class OperatorFactory:

    def __init__(self):
        self._factory = {}

    def register(self, name: str, op_class):
        """register an operator

        Args:
            name (str): user defined name of operator
            op_class (Operator): an Operator class

        Raise:
            KeyError if name is existed.
        """

        if name not in self._factory:
            self._factory[name] = op_class
        else:
            raise KeyError(f"{name} already exists.")

    def __getitem__(self, op_name):
        if op_name not in self._factory:
            raise KeyError(f"{op_name} is not registered")
        return self._factory[op_name]

    def __eq__(self, other):
        if not isinstance(other, OperatorFactory):
            return False
        return self._factory == other._factory



import pytest

from ...tests.common_tests import OperatorTestTemplate, ParamTuple
from ..custom_word_tokenizer import CustomWordTokenizer


@pytest.fixture(scope='module')
def user_words():
    return ['珍奶', '珍奶去冰', '去冰']


@pytest.fixture(scope='module')
def op(user_words):
    return CustomWordTokenizer(user_words)


class TestCustomWordTokenizer(OperatorTestTemplate):

    params = [
        ParamTuple(
            "珍奶",
            [1, 1],
            ["珍奶"],
            [1],
            id='exactly fit',
        ),
        ParamTuple(
            "去冰珍奶謝謝",
            [1, 1, 2, 2, 3, 4],
            ["去冰", "珍奶", "謝", "謝"],
            [1, 2, 3, 4],
            id='multiple',
        ),
        ParamTuple(
            "我想要珍奶去冰",
            [1, 2, 3, 4, 4, 4, 4],
            ["我", "想", "要", "珍奶去冰"],
            [1, 2, 3, 4],
            id='fit end',
        ),
        ParamTuple(
            "GB亂入",
            [1, 2, 3, 4],
            ["G", "B", "亂", "入"],
            [1, 2, 3, 4],
            id='not fit',
        ),
    ]

    def op(self):
        pass

    def test_equal(self, op, user_words):
        assert CustomWordTokenizer(user_words) == op

    @pytest.mark.parametrize(  # type: ignore
        "user_words",
        [
            pytest.param([], id='empty'),
            pytest.param(None, id='none'),
        ],
    )
    def test_invalid_init(self, user_words):
        with pytest.raises(ValueError):
            CustomWordTokenizer(user_words)


@pytest.mark.parametrize(
    "input_str,input_labels,expected_output_seq,"
    "expected_output_labels,expected_realigned_labels",
    [
        pytest.param(
            "我想要珍奶去冰",
            [1, 2, 3, 4, 4, 0, 4],
            ["我", "想", "要", "珍奶去冰"],
            [1, 2, 3, 4],
            [1, 2, 3, 4, 4, 4, 4],
            id='word has different labels',
        ),
    ],
)
def test_invertible_cases(
        input_str, input_labels,
        expected_output_seq, expected_output_labels,
        expected_realigned_labels,
        op,
    ):
    output_seq, label_aligner = op.transform(input_str)
    assert expected_output_seq == output_seq

    output_labels = label_aligner.transform(input_labels)
    assert expected_output_labels == output_labels

    realigned_labels = label_aligner.inverse_transform(expected_output_labels)
    assert expected_realigned_labels == realigned_labels



from unittest import TestCase

from ..elements import (
    Intent,
    Entity,
    Datum,
)
from ..exceptions import (
    EntityOverlapping,
    EntityPositionError,
)


class ElementIntentTestCase(TestCase):

    def setUp(self):
        self.intent_label = 1
        self.intent = Intent(self.intent_label)

    def test_init(self):
        self.assertEqual(self.intent.label, self.intent_label)

    def test_representation(self):
        self.assertEqual(f"<Intent {self.intent_label}>", self.intent.__repr__())


class ElementEntityTestCase(TestCase):

    def test_no_replacements(self):
        ent = Entity(
            label=1,
            value='紐約',
            start=1,
            end=3,
        )
        self.assertTrue(ent.no_replacements())
        self.assertEqual("<Entity 1: 紐約 at 1 - 3, with replacements: >", ent.__repr__())

    def test_has_replacements(self):
        ent = Entity(
            label=1,
            value='紐約',
            start=1,
            end=3,
            replacements=['台北'],
        )
        self.assertFalse(ent.no_replacements())
        self.assertEqual("<Entity 1: 紐約 at 1 - 3, with replacements: 台北>", ent.__repr__())

    def test_from_dict(self):
        label = 1
        start = 1
        end = 3
        replacements = ['斯堪地那維亞', 'KIX']
        utterance = '去紐約的機票'
        entity = {
            'label': label,
            'start': start,
            'end': end,
            'replacements': replacements,
        }
        ent = Entity.from_dict(entity, utterance)
        self.assertIsInstance(ent, Entity)
        self.assertEqual(ent.value, '紐約')
        self.assertEqual(ent.label, 1)
        self.assertEqual(ent.start, start)
        self.assertEqual(ent.end, end)
        self.assertEqual(ent.replacements, set(replacements))
        self.assertIn(
            ent.__repr__(),
            [
                "<Entity 1: 紐約 at 1 - 3, with replacements: 斯堪地那維亞, KIX>",
                "<Entity 1: 紐約 at 1 - 3, with replacements: KIX, 斯堪地那維亞>",
            ],
        )

    def test_from_dict_no_replacements(self):
        label = 1
        start = 1
        end = 3
        utterance = '去紐約的機票'
        entity = {
            'label': label,
            'start': start,
            'end': end,
        }
        ent = Entity.from_dict(entity, utterance)
        self.assertIsInstance(ent, Entity)
        self.assertEqual(ent.value, '紐約')
        self.assertEqual(ent.label, 1)
        self.assertEqual(ent.start, start)
        self.assertEqual(ent.end, end)
        self.assertEqual(ent.replacements, set([]))
        self.assertEqual("<Entity 1: 紐約 at 1 - 3, with replacements: >", ent.__repr__())

    def test_to_dict(self):
        label = 1
        start = 1
        end = 3
        replacements = ['斯堪地那維亞', 'KIX']
        entity = {
            'label': label,
            'start': start,
            'end': end,
            'replacements': replacements,
        }
        ent = Entity(
            label=label,
            value='紐約',
            start=start,
            end=end,
            replacements=replacements,
        )
        ent_tx = ent.to_dict()
        self.assertEqual(ent_tx['label'], entity['label'])
        self.assertEqual(ent_tx['start'], entity['start'])
        self.assertEqual(ent_tx['end'], entity['end'])
        self.assertEqual(set(ent_tx['replacements']),
                         set(entity['replacements']))

    def test_to_dict_no_replacement(self):
        label = 1
        start = 1
        end = 3
        entity = {
            'label': label,
            'start': start,
            'end': end,
        }
        ent = Entity(
            label=label,
            value='紐約',
            start=start,
            end=end,
        )
        ent_tx = ent.to_dict()
        self.assertEqual(ent_tx['label'], entity['label'])
        self.assertEqual(ent_tx['start'], entity['start'])
        self.assertEqual(ent_tx['end'], entity['end'])
        self.assertNotIn('replacements', ent_tx)

    def test_equal(self):
        ent1 = Entity(
            label=1,
            value='紐約',
            start=1,
            end=3,
            replacements=['台北'],
        )
        ent2 = Entity(
            label=1,
            value='紐約',
            start=1,
            end=3,
            replacements=['台北'],
        )
        self.assertEqual(ent1, ent2)

    def test_not_equal(self):
        ent = Entity(
            label=1,
            value='紐約',
            start=1,
            end=2,
            replacements=['台北'],
        )
        ent_diff_label = Entity(
            label=2,
            value='紐約',
            start=1,
            end=2,
            replacements=['台北'],
        )
        ent_diff_value = Entity(
            label=1,
            value='新加坡',
            start=1,
            end=2,
            replacements=['台北'],
        )
        ent_diff_start = Entity(
            label=1,
            value='紐約',
            start=0,
            end=2,
            replacements=['台北'],
        )
        ent_diff_end = Entity(
            label=1,
            value='紐約',
            start=1,
            end=3,
            replacements=['台北'],
        )
        ent_diff_replacements = Entity(
            label=1,
            value='紐約',
            start=1,
            end=2,
            replacements=['台北', 'KIX'],
        )
        self.assertNotEqual(ent, ent_diff_label)
        self.assertNotEqual(ent, ent_diff_value)
        self.assertNotEqual(ent, ent_diff_start)
        self.assertNotEqual(ent, ent_diff_end)
        self.assertNotEqual(ent, ent_diff_replacements)

    def test_raise_error_if_compare_different_type(self):
        ent = Entity(
            label=1,
            value='紐約',
            start=1,
            end=2,
            replacements=['台北'],
        )
        ent_diff_type = object()
        with self.assertRaises(TypeError):
            ent == ent_diff_type


class ElementDatumTestCase(TestCase):

    def test_raise_entity_position_error_if_entity_position_wrong(self):
        with self.subTest(case='wrong end'):
            with self.assertRaises(EntityPositionError):
                Datum(
                    utterance='去紐約的機票',
                    intents=[Intent(0)],
                    entities=[
                        Entity(label=1, value='紐約', start=1, end=2),
                    ],
                )
        with self.subTest(case='wrong start'):
            with self.assertRaises(EntityPositionError):
                Datum(
                    utterance='去紐約的機票',
                    intents=[Intent(0)],
                    entities=[
                        Entity(label=1, value='紐約', start=0, end=3),
                    ],
                )
        with self.subTest(case='wrong value'):
            with self.assertRaises(EntityPositionError):
                Datum(
                    utterance='去紐曰的機票',
                    intents=[Intent(0)],
                    entities=[
                        Entity(label=1, value='紐約', start=1, end=3),
                    ],
                )

    def test_raise_entity_overlap_if_entity_overlaps(self):
        with self.assertRaises(EntityOverlapping):
            Datum(
                utterance='我想吃麥當勞快樂兒童餐',
                intents=[Intent(0)],
                entities=[
                    Entity(label=0, value='快樂', start=6, end=8),
                    Entity(label=1, value='快樂兒童餐', start=6, end=11),
                ],
            )

    def test_entities_correct_order_in_datum(self):
        utterance = '我想訂明天從紐約飛到新加坡的機票'
        entities = [
            Entity(label=0, value='新加坡', start=10,
                   end=13, replacements=['斯堪地那維亞', 'KIX']),
            Entity(label=1, value='明天', start=3,
                   end=5, replacements=['下禮拜二']),
            Entity(label=2, value='紐約', start=6, end=8),
        ]
        intents = [
            Intent(1),
        ]
        datum = Datum(
            utterance=utterance,
            intents=intents,
            entities=entities,
        )

        expected_entities = [
            Entity(label=1, value='明天', start=3,
                   end=5, replacements=['下禮拜二']),
            Entity(label=2, value='紐約', start=6, end=8),
            Entity(label=0, value='新加坡', start=10,
                   end=13, replacements=['斯堪地那維亞', 'KIX']),
        ]
        for result_entity, expected_entity in zip(datum.entities, expected_entities):
            self.assertEqual(result_entity.label, expected_entity.label)
            self.assertEqual(result_entity.value, expected_entity.value)
            self.assertEqual(result_entity.start, expected_entity.start)
            self.assertEqual(result_entity.end, expected_entity.end)

    def test_has_entities(self):
        datum = Datum(
            utterance='去紐約的機票',
            intents=[Intent(0)],
            entities=[
                Entity(label=1, value='紐約', start=1, end=3),
            ],
        )
        self.assertTrue(datum.has_entities())

    def test_has_no_entities(self):
        datum = Datum(
            utterance='去紐約的機票',
            intents=[Intent(0)],
            entities=[],
        )
        self.assertFalse(datum.has_entities())

    def test_has_intents(self):
        datum = Datum(
            utterance='去紐約的機票',
            intents=[Intent(0)],
            entities=[
                Entity(label=1, value='紐約', start=1, end=3),
            ],
        )
        self.assertTrue(datum.has_intents())

    def test_has_no_intents(self):
        datum = Datum(
            utterance='去紐約的機票',
            intents=None,
            entities=[
                Entity(label=1, value='紐約', start=1, end=3),
            ],
        )
        self.assertFalse(datum.has_intents())

    def test_same_utterance(self):
        utterance = '我想訂明天從紐約飛到新加坡的機票'
        entities = [
            Entity(label=0, value='新加坡', start=10,
                   end=13, replacements=['斯堪地那維亞', 'KIX']),
            Entity(label=1, value='明天', start=3,
                   end=5, replacements=['下禮拜二']),
            Entity(label=2, value='紐約', start=6, end=8),
        ]
        intents = [
            Intent(1),
        ]
        datum = Datum(
            utterance=utterance,
            intents=intents,
            entities=entities,
        )
        datum2 = Datum(
            utterance=utterance,
        )
        self.assertTrue(datum.has_same_utterance_as(datum2))

    def test_same_intents(self):
        utterance = '我想訂明天從紐約飛到新加坡的機票'
        utterance2 = '我想訂機票'
        entities = [
            Entity(label=0, value='新加坡', start=10,
                   end=13, replacements=['斯堪地那維亞', 'KIX']),
            Entity(label=1, value='明天', start=3,
                   end=5, replacements=['下禮拜二']),
            Entity(label=2, value='紐約', start=6, end=8),
        ]
        intents = [
            Intent(1),
        ]
        intents2 = [
            Intent(1),
        ]
        datum = Datum(
            utterance=utterance,
            intents=intents,
            entities=entities,
        )
        datum2 = Datum(
            utterance=utterance2,
            intents=intents2,
        )
        self.assertTrue(datum.has_same_intents_as(datum2))

    def test_same_entities(self):
        utterance = '我想訂明天從紐約飛到新加坡的機票'
        utterance2 = '我想查明天從紐約飛往新加坡的機票'
        entities = [
            Entity(label=0, value='新加坡', start=10,
                   end=13, replacements=['斯堪地那維亞', 'KIX']),
            Entity(label=1, value='明天', start=3,
                   end=5, replacements=['下禮拜二']),
            Entity(label=2, value='紐約', start=6, end=8),
        ]
        entities2 = [
            Entity(label=0, value='新加坡', start=10,
                   end=13, replacements=['斯堪地那維亞', 'KIX']),
            Entity(label=1, value='明天', start=3,
                   end=5, replacements=['下禮拜二']),
            Entity(label=2, value='紐約', start=6, end=8),
        ]
        intents = [
            Intent(1),
        ]
        intents2 = [
            Intent(0),
        ]
        datum = Datum(
            utterance=utterance,
            intents=intents,
            entities=entities,
        )
        datum2 = Datum(
            utterance=utterance2,
            intents=intents2,
            entities=entities2,
        )
        self.assertTrue(datum.has_same_entities_as(datum2))



import pytest

from ..utils import unpack_datum, pack_to_datum
from uttut.elements import Datum, Intent, Entity
from uttut import ENTITY_LABEL

not_entity = ENTITY_LABEL['NOT_ENTITY']


def test_unpack_datum():
    utterance = '我想訂明天從紐約飛到新加坡的機票'
    entities = [
        Entity(label=1, value='新加坡', start=10,
               end=13, replacements=['斯堪地那維亞', 'KIX']),
        Entity(label=2, value='明天', start=3,
               end=5, replacements=['下禮拜二']),
        Entity(label=3, value='紐約', start=6, end=8),
    ]
    intents = [Intent(1)]
    datum = Datum(
        utterance=utterance,
        intents=intents,
        entities=entities,
    )
    output_utt, output_intent, output_entity = unpack_datum(datum)

    assert utterance == output_utt
    assert [1] == output_intent
    assert [0, 0, 0, 2, 2, 0, 3, 3, 0, 0, 1, 1, 1, 0, 0, 0] == output_entity


@pytest.mark.parametrize(
    "input_tuple,expected_datum",
    [
        pytest.param(
            ("abb", [1, 2], [1, 2, 2]),
            Datum(
                "abb",
                [Intent(1), Intent(2)],
                [Entity(1, "a", 0, 1), Entity(2, "bb", 1, 3)],
            ),
            id='all entities',
        ),
        pytest.param(
            ("acbb", [1, 2], [1, not_entity, 2, 2]),
            Datum(
                "acbb",
                [Intent(1), Intent(2)],
                [Entity(1, "a", 0, 1), Entity(2, "bb", 2, 4)],
            ),
            id='not entity in the middle',
        ),
        pytest.param(
            (" acbb", [1, 2], [not_entity, 1, not_entity, 2, 2]),
            Datum(
                " acbb",
                [Intent(1), Intent(2)],
                [Entity(1, "a", 1, 2), Entity(2, "bb", 3, 5)],
            ),
            id='not entity in head',
        ),
        pytest.param(
            (" acbbo", [1, 2], [not_entity, 1, not_entity, 2, 2, not_entity]),
            Datum(
                " acbbo",
                [Intent(1), Intent(2)],
                [Entity(1, "a", 1, 2), Entity(2, "bb", 3, 5)],
            ),
            id='not entity in tail',
        ),
        pytest.param(
            ("abb", [1, 2, 2, 1, 1, 1], [1, 2, 2]),
            Datum(
                "abb",
                [Intent(1), Intent(2)],
                [Entity(1, "a", 0, 1), Entity(2, "bb", 1, 3)],
            ),
            id='duplicated intent labels',
        ),
        pytest.param(
            ("abc", [1, 2], [1, 2, 3]),
            Datum(
                "abc",
                [Intent(1), Intent(2)],
                [Entity(1, "a", 0, 1), Entity(2, "b", 1, 2), Entity(3, "c", 2, 3)],
            ),
            id='char-level all different',
        ),
    ],
)
def test_pack_to_datum(input_tuple, expected_datum):
    output = pack_to_datum(*input_tuple)
    assert expected_datum == output


def test_pack_to_datum_fail():
    with pytest.raises(ValueError):
        pack_to_datum('abb', [1, 2], [1, 3])



from typing import List, Iterable
import pprint
cimport cython  # noqa: E999

from .exceptions import (
    EntityOverlapping,
    EntityPositionError,
)
from .utils import (
    entity_position_correct,
    overlap,
    msg_entity_wrong_position,
    msg_entity_overlapping,
)


cdef class Entity:

    cdef public int label
    cdef public str value
    cdef public int start
    cdef public int end
    cdef public set replacements
    cdef public int index

    def __init__(
            self,
            int label,
            str value,
            int start,
            int end,
            replacements: Iterable[str] = None,
        ) -> None:
        self.index = -1

        self.label = label
        self.value = value
        self.start = start
        self.end = end
        cdef list _replacements = [] if replacements is None else list(replacements)
        self.replacements = set(_replacements)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError('can only compare entity to entity')
        cdef bint same_label = self.label == other.label
        cdef bint same_value = self.value == other.value
        cdef bint same_position = (self.start == other.start) and (self.end == other.end)
        cdef bint same_replacements = self.replacements == other.replacements
        if same_label and same_value and same_position and same_replacements:
            return True
        return False

    def __repr__(self):
        return "<Entity {}: {} at {} - {}, with replacements: {}>".format(
            self.label,
            self.value,
            self.start,
            self.end,
            ', '.join(list(self.replacements)),
        )

    def no_replacements(self) -> bool:
        return len(self.replacements) == 0

    def n_replacements(self):
        return len(self.replacements) + 1

    def to_dict(self) -> dict:
        result = {
            'label': self.label,
            'start': self.start,
            'end': self.end,
        }
        if not self.no_replacements():
            result['replacements'] = list(self.replacements)
        return result

    @classmethod
    def from_dict(cls, dict entity, str utterance):
        new_end = entity['end']
        entity_value = utterance[entity['start']: new_end]
        replacements = entity.get('replacements')
        return cls(
            label=entity['label'],
            value=entity_value,
            start=entity['start'],
            end=new_end,
            replacements=[] if replacements is None else replacements,
        )


cdef class Intent:

    cdef public int label

    def __init__(self, int label):
        self.label = label

    def __repr__(self):
        return f"<Intent {self.label}>"

    def __hash__(self):
        return hash(self.label)

    def __eq__(self, other):
        return self.label == other.label


cdef class Datum:

    cdef public str utterance
    cdef public list intents
    cdef public list entities

    def __init__(
            self,
            str utterance,
            list intents=None,  # : List[Intent] = None,
            list entities=None,  # : List[Entity] = None,
        ):
        self.utterance = utterance
        self.intents = [] if intents is None else sorted(intents, key=lambda i: hash(i))
        self.entities = [] if entities is None else sorted(entities, key=lambda e: e.start)

        # check entity has correct value and position
        for entity in self.entities:
            if not entity_position_correct(self.utterance, entity):
                raise EntityPositionError(msg_entity_wrong_position(self.utterance, entity))

        # check overlapping entities
        for idx in range(len(self.entities) - 1):
            entity = self.entities[idx]
            next_entity = self.entities[idx + 1]
            if overlap(entity, next_entity):
                raise EntityOverlapping(msg_entity_overlapping(self.utterance, entity, next_entity))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError('can only compare datum to datum')

        if not self.has_same_utterance_as(other):
            return False
        elif not self.has_same_intents_as(other):
            return False
        elif not self.has_same_entities_as(other):
            return False
        else:
            return True

    def __repr__(self):
        intent_str = pprint.pformat(self.intents)
        entity_str = pprint.pformat(self.entities)
        return f"<Datum {self.utterance}\n with intents: "\
            f"{intent_str}\n with entities:\n{entity_str}>"

    def has_same_utterance_as(self, other):
        return self.utterance == other.utterance

    def has_same_intents_as(self, other):
        return set(self.intents) == set(other.intents)

    def has_same_entities_as(self, other):
        same_entities = False
        if len(self.entities) == len(other.entities):
            for self_entity, other_entity in zip(self.entities, other.entities):
                if self_entity != other_entity:
                    break
            else:
                same_entities = True
        return same_entities

    def has_entities(self):
        return len(self.entities) > 0

    def has_intents(self):
        return len(self.intents) > 0

    def copy_intents(self) -> List[Intent]:
        return [Intent(intent.label) for intent in self.intents]



import pytest

from ..replacement import ReplacementGroup
from uttut.pipeline.edit import str2str


test_cases = [
    pytest.param(
        '我想要喝200元的珍奶10杯',
        ReplacementGroup.add_all([(4, 7, '_int_'), (11, 13, '_int_')]),
        '我想要喝_int_元的珍奶_int_杯',
        ReplacementGroup.add_all([(4, 9, '200'), (13, 18, '10')]),
        id='modify',
    ),
    pytest.param(
        '我想喝珍奶',
        ReplacementGroup.add_all([(0, 0, '其實'), (0, 1, '妳'), (1, 1, '不'),
                                  (3, 5, '多多綠'), (5, 5, '對吧！')]),
        '其實妳不想喝多多綠對吧！',
        ReplacementGroup.add_all([(0, 2, ''), (2, 3, '我'), (3, 4, ''),
                                  (6, 9, '珍奶'), (9, 12, '')]),
        id='insert',
    ),
    pytest.param(
        '我不想喝珍奶',
        ReplacementGroup.add_all([(1, 2, '')]),
        '我想喝珍奶',
        ReplacementGroup.add_all([(1, 1, '不')]),
        id='delete',
    ),
    pytest.param(
        '這個句子沒有改變',
        ReplacementGroup.add_all([]),
        '這個句子沒有改變',
        ReplacementGroup.add_all([]),
        id='identity',
    ),
]


@pytest.mark.parametrize(
    "input_str,forward_replacements,output_str,inverse_replacements",
    test_cases,
)
def test_forward_apply(input_str, forward_replacements, output_str, inverse_replacements):
    output = str2str.apply(input_str, forward_replacements)
    assert output_str == output


@pytest.mark.parametrize(
    "input_str,forward_replacements,output_str,inverse_replacements",
    test_cases,
)
def test_backward_apply(input_str, forward_replacements, output_str, inverse_replacements):
    output = str2str.apply(output_str, inverse_replacements)
    assert input_str == output


@pytest.mark.parametrize(
    "input_str,forward_replacements,output_str,inverse_replacements",
    test_cases,
)
def test_forward_inverse(input_str, forward_replacements, output_str, inverse_replacements):
    output = str2str.inverse(input_str, forward_replacements)
    assert inverse_replacements == output


@pytest.mark.parametrize(
    "input_str,forward_replacements,output_str,inverse_replacements",
    test_cases,
)
def test_backward_inverse(input_str, forward_replacements, output_str, inverse_replacements):
    output = str2str.inverse(output_str, inverse_replacements)
    assert forward_replacements == output



import pytest

from .tokenization import WordpieceTokenizer as BertWordpieceTokenizer
from ..word_piece import word_piece_pipe, vocab

from uttut.elements import Datum


test_cases = [
    pytest.param(  # ref: Bert tokenization_test.py#L87
        "unwanted running",
        id="eng",
    ),
    pytest.param(  # ref: Bert tokenization_test.py#L91
        "unwantedX running",
        id="eng with unk",
    ),
]


@pytest.fixture
def tokenizer():
    yield BertWordpieceTokenizer(vocab)


@pytest.mark.parametrize("input_str", test_cases)
def test_all(input_str, tokenizer):
    datum = Datum(input_str)
    output, _, _, _, _ = word_piece_pipe.transform(datum)
    expected_output = tokenizer.tokenize(input_str)
    assert expected_output == output



from typing import List

from .replacement import ReplacementGroup
from .span import SpanGroup


def apply(input_str: str, span_group: SpanGroup) -> List[str]:

    _validate_compatibility(input_str, span_group)

    output_lst = [''] * len(span_group)
    for i, span in enumerate(span_group):
        output_lst[i] = input_str[span.start: span.end]

    return output_lst


def _validate_compatibility(input_str: str, span_group: SpanGroup):
    if len(input_str) != span_group[-1].end:
        raise ValueError('Input list and span group is not compatible.')


def gen_span_group(input_str: str, tokens: List[str]) -> SpanGroup:
    '''Compare string and tokens then generate SpanGroup'''

    if input_str != ''.join(tokens):
        raise ValueError('input_str and tokens are not compatible')

    shift = 0
    span_group = SpanGroup()
    for token in tokens:
        start = input_str.find(token, shift)
        end = start + len(token)
        span_group.add(start, end)
        shift = end
    span_group.done()
    assert len(span_group) == len(tokens)
    return span_group


def gen_replacement_group(input_str: str, tokens: List[str]) -> ReplacementGroup:
    '''Compare string and tokens then generate ReplacementGroup'''

    shift = 0
    replacement_group = ReplacementGroup()
    for token in tokens:
        start = input_str.find(token, shift)
        if start == -1:
            raise ValueError('input_str and tokens are not compatible')
        if input_str[shift: start] != '':
            replacement_group.add(shift, start, '')
        shift = start + len(token)

    if shift != len(input_str):
        replacement_group.add(shift, len(input_str), '')

    replacement_group.done()
    return replacement_group



import pytest

from ..span_subwords import SpanSubwords
from .common_tests import OperatorTestTemplate, ParamTuple


def to_dict(input_list):
    output_dict = {}
    for i, token in enumerate(input_list):
        output_dict[token] = i
    return output_dict


class TestSpanSubwordsNoSubwords(OperatorTestTemplate):

    params = [
        ParamTuple(
            ["I", "like", "apples"],
            [1, 2, 3],
            ["I", "like", "apples"],
            [1, 2, 3],
            id="exactly match",
        ),
        ParamTuple(
            ["I", "like", "bananas"],
            [1, 2, 3],
            ["I", "like", "<unk>"],
            [1, 2, 3],
            id="word not found",
        ),
        ParamTuple(
            ["Ilikebananas"],
            [3],
            ["<unk>"],
            [3],
            id="word too long",
        ),
    ]

    @pytest.fixture(scope='class')
    def op(self):
        return SpanSubwords(
            vocab=to_dict(["I", "like", "apples"]),
            unk_token="<unk>",
            maxlen_per_token=10,
        )


class TestSpanSubwordsHasSubwords(OperatorTestTemplate):

    params = [
        ParamTuple(
            ["unwanted", "running"],
            [1, 2],
            ["un", "##want", "##ed", "runn", "##ing"],
            [1, 1, 1, 2, 2],
            id="exactly match",
        ),
        ParamTuple(
            ["unwantedX", "running"],
            [1, 2],
            ["[UNK]", "runn", "##ing"],
            [1, 2, 2],
            id="word not found",
        ),
    ]

    @pytest.fixture(scope='class')
    def op(self):
        return SpanSubwords(
            vocab=to_dict(["[UNK]", "[CLS]", "[SEP]", "want",
                           "##want", "##ed", "wa", "un", "runn", "##ing"]),
            unk_token="[UNK]",
            maxlen_per_token=200,
        )


@pytest.mark.parametrize(
    "obj1,obj2",
    [
        pytest.param(
            SpanSubwords(
                vocab=to_dict(["I", "like", "apples"]),
                unk_token="<unk>",
                maxlen_per_token=10,
            ),
            SpanSubwords(
                vocab=to_dict(["I", "like", "apples"]),
                unk_token="<unk>",
                maxlen_per_token=200,
            ),
            id="different maxlen_per_token",
        ),
        pytest.param(
            SpanSubwords(
                vocab=to_dict(["I", "like", "apples"]),
                unk_token="<unk>",
                maxlen_per_token=10,
            ),
            SpanSubwords(
                vocab=to_dict(["I", "like", "apples"]),
                unk_token="[UNK]",
                maxlen_per_token=10,
            ),
            id="different unk",
        ),
        pytest.param(
            SpanSubwords(
                vocab=to_dict(["I", "like", "apples"]),
                unk_token="<unk>",
                maxlen_per_token=10,
            ),
            SpanSubwords(
                vocab=to_dict(["##apples"]),
                unk_token="<unk>",
                maxlen_per_token=10,
            ),
            id="different vocab",
        ),
    ],
)
def test_not_equal(obj1, obj2):
    assert obj1 != obj2



import pytest

from ...tests.common_tests import OperatorTestTemplate, ParamTuple
from ..zh_char_tokenizer import ZhCharTokenizer


class TestZhCharTokenizer(OperatorTestTemplate):

    params = [
        ParamTuple(
            "How's it going today, Mr.Smith?",
            [1, 1, 1, 1, 1, 0, 2, 2, 0, 3, 3, 3, 3, 3, 0, 4, 4, 4, 4, 4, 5, 0,
             6, 6, 6, 6, 6, 6, 6, 6, 7],
            ['How', "'", 's', 'it', 'going', 'today', ',', 'Mr', '.', 'Smith', '?'],
            [1, 1, 1, 2, 3, 4, 5, 6, 6, 6, 7],
            id='eng',
        ),
        ParamTuple(
            "薄餡的櫻花妹呢？",
            [1, 1, 0, 2, 2, 2, 3, 4],
            ['薄', '餡', '的', '櫻', '花', '妹', '呢', '？'],
            [1, 1, 0, 2, 2, 2, 3, 4],
            id='zh',
        ),
        ParamTuple(
            "GB亂入",
            [1, 1, 2, 2],
            ['GB', '亂', '入'],
            [1, 2, 2],
            id='eng + zh',
        ),
        ParamTuple(
            "GB亂入!!!",
            [1, 1, 2, 2, 3, 4, 5],
            ['GB', '亂', '入', '!', '!', '!'],
            [1, 2, 2, 3, 4, 5],
            id='eng + zh + punct',
        ),
    ]

    @pytest.fixture(scope='class')
    def op(self):
        return ZhCharTokenizer()

    def test_equal(self, op):
        assert op == ZhCharTokenizer()



import pytest

from ..add_whitespace_around_cjk import AddWhitespaceAroundCJK
from .common_tests import OperatorTestTemplate, ParamTuple


class TestAddWhitespaceAroundCJK(OperatorTestTemplate):

    params = [
        ParamTuple(
            "alvin喜歡吃榴槤",
            [1, 1, 1, 1, 1, 2, 3, 4, 5, 6],
            "alvin 喜  歡  吃  榴  槤 ",
            [1, 1, 1, 1, 1, 0, 2, 0, 0, 3, 0, 0, 4, 0, 0, 5, 0, 0, 6, 0],
            'eng + zh',
        ),
        ParamTuple(
            "hello! how are you?",
            [1, 1, 1, 1, 1, 2, 0, 3, 3, 3, 0, 4, 4, 4, 0, 5, 5, 5, 6],
            "hello! how are you?",
            [1, 1, 1, 1, 1, 2, 0, 3, 3, 3, 0, 4, 4, 4, 0, 5, 5, 5, 6],
            'eng',
        ),
    ]

    @pytest.fixture(scope='class')
    def op(self):
        return AddWhitespaceAroundCJK()

    def test_equal(self, op):
        assert op == AddWhitespaceAroundCJK()



# -*- coding: utf-8 -*-
from unittest import TestCase

from uttut.elements import (
    Datum,
    Intent,
    Entity,
)
from ..get_unique_entities import (
    get_unique_entities,
)


class GetUniqueEntitiesTestCase(TestCase):

    def setUp(self):
        self.intents = [Intent(0)]
        self.entities = [
            Entity(label=1, value='豪大', start=7, end=9,
                   replacements=['肯德基', '麥噹噹']),
            Entity(label=2, value='雞排', start=9, end=11,
                   replacements=['漢堡', '薯條', '滷肉飯']),
        ]
        self.example_datum = Datum(
            utterance='家豪大大喜歡吃豪大雞排',
            intents=self.intents,
            entities=self.entities,
        )

    def test_get_unique_entities(self):
        result = get_unique_entities(data=[self.example_datum])
        self.assertEqual(set([1, 2]), set(result))



from typing import List, Tuple

from uttut.elements import Datum, Intent, Entity
from uttut import ENTITY_LABEL


def unpack_datum(datum: Datum) -> Tuple[str, List[int], List[int]]:

    utterance = datum.utterance
    uttlen = len(utterance)

    # intent
    intent_labels = [intent.label for intent in datum.intents]

    # entity
    entity_labels = [ENTITY_LABEL['NOT_ENTITY']] * uttlen
    for entity in datum.entities:
        label = entity.label
        for ind in range(entity.start, entity.end):
            entity_labels[ind] = label

    return utterance, intent_labels, entity_labels


def pack_to_datum(utterance: str, intent_labels: List[int], entity_labels: List[int]) -> Datum:

    if len(utterance) != len(entity_labels):
        raise ValueError('utterance and entity_labels are not compatible')

    # intent
    intents = [Intent(intent_label) for intent_label in sorted(set(intent_labels))]

    # entity
    entities = []
    start_indices = _get_meaningful_start_indices(entity_labels)
    end_indices = _get_meaningful_start_indices(entity_labels[::-1])
    for start, end in zip(start_indices, end_indices[::-1]):
        real_end = len(utterance) - end
        entity = Entity(
            start=start,
            end=real_end,
            value=utterance[start: real_end],
            label=entity_labels[start],
        )
        entities.append(entity)

    # datum
    return Datum(utterance=utterance, intents=intents, entities=entities)


def _get_meaningful_start_indices(labels: List[int]) -> List[int]:
    indices = _get_start_indices(labels)
    j = 0
    for ind in indices:
        if labels[ind] != ENTITY_LABEL['NOT_ENTITY']:
            indices[j] = ind
            j += 1
    return indices[: j]


def _get_start_indices(labels: List[int]) -> List[int]:
    current_label = labels[0]
    indices = [0]
    for i, label in enumerate(labels):
        if label != current_label:
            indices.append(i)
            current_label = label
    print(indices)
    return indices



from typing import List
import re
from collections import Counter

from uttut import ENTITY_LABEL
from ..label_transducer import get_most_common_except_not_entity
from ..tokens import FLOAT_TOKEN_WITH_SPACE
from .base import PatternRecognizer, PatternRecognizerAligner


class FloatTokenWithSpaceAligner(PatternRecognizerAligner):

    def _forward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        token = FloatTokenWithSpace.TOKEN
        counter = Counter(labels)
        common_label = counter.most_common()[0][0]
        token_len = len(token) - 2
        output_label = [ENTITY_LABEL['NOT_ENTITY']] + \
            [common_label] * token_len + [ENTITY_LABEL['NOT_ENTITY']]
        return output_label

    def _backward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_most_common_except_not_entity(labels, output_size)


class FloatTokenWithSpace(PatternRecognizer):
    """
    Recognize integer (ex: 12.3, 100.77) in the input string
    and replace them with FLOAT_TOKEN_WTIH_SPACE ( _float_ )

    Note that extra spaces are added behind and after the FLOAT_TOKEN.

    E.g.
    >>> from uttut.pipeline.ops.float_token_with_space import FloatTokenWithSpace
    >>> op = FloatTokenWithSpace()
    >>> output_seq, label_aligner = op.transform("10.7")
    >>> output_labels = label_aligner.transform([1, 1, 1, 1])
    >>> output_seq
    " _float_ "
    >>> output_labels
    [0, 1, 1, 1, 1, 1, 1, 1, 0]
    >>> label_aligner.inverse_transform(output_labels)
    [1, 1, 1, 1]

    """

    _label_aligner_class = FloatTokenWithSpaceAligner

    REGEX_PATTERN = re.compile(r"(?<![\.\d])\d+\.\d+(?![\.\d])")
    TOKEN = FLOAT_TOKEN_WITH_SPACE

    def _gen_forward_replacement_group(self, input_str: str):  # type: ignore
        return super()._gen_forward_replacement_group(
            input_str=input_str,
            annotation='float-token-with-space',
        )



class UttutBaseException(Exception):
    pass


class EntityPositionError(UttutBaseException):
    pass


class EntityOverlapping(UttutBaseException):
    pass


class DifferentUtterance(UttutBaseException):
    pass



from typing import List, Tuple
import unicodedata

from .base import Operator, LabelAligner
from .label_transducer import get_most_common_except_not_entity
from ..edit.replacement import ReplacementGroup
from ..edit import str2str
from ..edit.label_propagation import propagate_by_replacement_group


class StripAccentToken(Operator):

    """
    Strip accent token

    E.g.
    >>> from uttut.pipeline.ops.strip_accent_token import StripAccentToken
    >>> op = StripAccentToken()
    >>> output_seq, label_aligner = op.transform(u"H\u00E9llo")
    >>> output_labels = label_aligner.transform([1, 2, 3, 4, 5])
    >>> output_seq
    "Hello"
    >>> output_labels
    [1, 2, 3, 4, 5]
    >>> label_aligner.inverse_transform(output_labels)
    [1, 2, 3, 4, 5]

    """

    _input_type = str
    _output_type = str

    def _transform(self, input_sequence: str) -> Tuple[str, 'LabelAligner']:
        forward_replacement_group = self._gen_forward_replacement_group(input_sequence)
        output_sequence = str2str.apply(input_sequence, forward_replacement_group)

        label_aligner = StripAccentTokenAligner(
            input_sequence=input_sequence,
            edit=forward_replacement_group,
            output_length=len(output_sequence),
        )
        return output_sequence, label_aligner

    def _gen_forward_replacement_group(self, input_str: str) -> ReplacementGroup:
        replacement_group = ReplacementGroup()
        output_str = _strip_accents(input_str)
        assert len(input_str) == len(output_str)

        for i, (input_char, output_char) in enumerate(zip(input_str, output_str)):
            if input_char != output_char:
                replacement_group.add(
                    start=i,
                    end=i + 1,
                    new_value=output_char,
                    annotation="strip-accent-token",
                )
        replacement_group.done()
        return replacement_group


class StripAccentTokenAligner(LabelAligner):

    def _transform(self, labels: List[int]) -> List[int]:
        return propagate_by_replacement_group(
            labels=labels,
            replacement_group=self._forward_edit,
            transduce_func=self._forward_transduce_func,
        )

    def _forward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_most_common_except_not_entity(labels, output_size)

    def _inverse_transform(self, labels: List[int]) -> List[int]:
        inverse_replacement_group = str2str.inverse(self._input_sequence, self._forward_edit)
        return propagate_by_replacement_group(
            labels=labels,
            replacement_group=inverse_replacement_group,
            transduce_func=self._backward_transduce_func,
        )

    def _backward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_most_common_except_not_entity(labels, output_size)


def _strip_accents(text: str) -> str:

    """Strips accents from a piece of text

    This code is copied from Bert `tokenization.py`.

    """

    text = unicodedata.normalize("NFD", text)
    output = []
    for char in text:
        cat = unicodedata.category(char)
        if cat == "Mn":
            continue
        output.append(char)
    return "".join(output)



def _validate_start_end(start, end):
    _validate_start_end_in_c(start, end)


def _validate_disjoint(sorted_objs):
    _validate_disjoint_in_c(sorted_objs)


cdef void _validate_start_end_in_c(unsigned int start, unsigned int end) except *:  # noqa: E999
    if start > end:
        raise ValueError("start cannot be greater than end")


cdef void _validate_disjoint_in_c(list sorted_objs) except *:
    '''
    An object must have two integer properties `start` and `end`.
    '''
    cdef unsigned int current

    current = sorted_objs[0].end  # type: ignore
    for obj in sorted_objs[1:]:
        if obj.start < current:  # type: ignore
            raise ValueError(f"overlapped")
        current = obj.end  # type: ignore



from typing import List

from ..label_transducer import get_most_common_except_not_entity
from .base import Tokenizer, TokenizerAligner


class WhiteSpaceTokenizerAligner(TokenizerAligner):

    def _forward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_most_common_except_not_entity(labels, output_size)

    def _backward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_most_common_except_not_entity(labels, output_size)


class WhiteSpaceTokenizer(Tokenizer):

    """
    Spilt input string into pieces by whitespace characters

    This is a modification of `whitespace_tokenize` in Bert `tokenization.py`.

    E.g.
    >>> from uttut.pipeline.ops.whitespace_tokenizer import WhiteSpaceTokenizer
    >>> op = WhiteSpaceTokenizer()
    >>> output_seq, label_aligner = op.transform("a b")
    >>> output_labels = label_aligner.transform([1, 2, 3])
    >>> output_seq
    ["a", "b"]
    >>> output_labels
    [1, 3]
    >>> label_aligner.inverse_transform(output_labels)
    [1, 0, 3]

    """

    _label_aligner_class = WhiteSpaceTokenizerAligner

    def _tokenize(self, input_str: str) -> List[str]:
        """Split text into list by whitespace characters

        E.g.
        1. "a  \t\t\n\n b" -> ["a", "b"]
        2. "  \n\t\n  " -> []
        3. "  a b \t\n" -> ["a", "b"]

        """
        return input_str.split()



import pytest

from ..token_to_index_with_hash import Token2IndexwithHash
from ..utils.consistent_hash import consistent_hash
from .common_tests import OperatorTestTemplate, ParamTuple


def hash_unknown(key):
    return consistent_hash(key, 5)


class TestToken2IndexwithHash(OperatorTestTemplate):

    params = [
        ParamTuple(
            ['薄餡', '要', '帶', '櫻花', '妹', '回來'],
            [1, 2, 3, 4, 5, 6],
            [1, 2, 3, hash_unknown('櫻花'), 4, hash_unknown('回來')],
            [1, 2, 3, 4, 5, 6],
            id='zh',
        ),
    ]

    @pytest.fixture(scope='class')
    def token2index(self):
        return {'oh': 0, '薄餡': 1, '要': 2, '帶': 3, '妹': 4}

    @pytest.fixture(scope='class')
    def op(self, token2index):
        return Token2IndexwithHash(token2index)

    @pytest.mark.parametrize(
        'token2index',
        [
            pytest.param(
                {'薄餡': 1, '要': 1, '帶': 3, '妹': 4},
                id='duplicated indices',
            ),
            pytest.param(
                {'薄餡': 0, '要': 2, '帶': 3, '妹': 4},
                id='indices not continuous',
            ),
        ],
    )
    def test_invalid_token2index(self, token2index):
        with pytest.raises(ValueError):
            Token2IndexwithHash(token2index=token2index)

    def test_equal(self, op, token2index):
        assert op == Token2IndexwithHash(token2index)

    def test_not_equal(self, op):
        op1 = Token2IndexwithHash(token2index={'薄餡': 0, '要': 1, '帶': 2, '妹': 3})
        assert op != op1



from typing import List

import re
from collections import Counter

from uttut import ENTITY_LABEL
from ..label_transducer import get_most_common_except_not_entity
from ..tokens import INT_TOKEN_WITH_SPACE
from .base import PatternRecognizer, PatternRecognizerAligner


class IntTokenWithSpaceAligner(PatternRecognizerAligner):

    def _forward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        token = IntTokenWithSpace.TOKEN
        counter = Counter(labels)
        common_label = counter.most_common()[0][0]
        token_len = len(token) - 2
        output_label = [ENTITY_LABEL['NOT_ENTITY']] + \
            [common_label] * token_len + [ENTITY_LABEL['NOT_ENTITY']]
        return output_label

    def _backward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_most_common_except_not_entity(labels, output_size)


class IntTokenWithSpace(PatternRecognizer):

    """
    Recognize integer (ex: 12, 10000) in the input string
    and replace them with INT_TOKEN_WTIH_SPACE ( _int_ )

    E.g.
    >>> from uttut.pipeline.ops.int_token_with_space import IntTokenWithSpace
    >>> op = IntTokenWithSpace()
    >>> output_seq, label_aligner = op.transform("10")
    >>> output_labels = label_aligner.transform([1, 1])
    >>> output_seq
    " _int_ "
    >>> output_labels
    [0, 1, 1, 1, 1, 1, 0]
    >>> label_aligner.inverse_transform(output_labels)
    [1, 1]

    """

    REGEX_PATTERN = re.compile(r"(?<![\.\d])\d+(?![\.\d])")
    TOKEN = INT_TOKEN_WITH_SPACE
    _label_aligner_class = IntTokenWithSpaceAligner

    def _gen_forward_replacement_group(self, input_str: str):  # type: ignore
        return super()._gen_forward_replacement_group(
            input_str=input_str,
            annotation='int-token-with-space',
        )



import pytest

from ..replacement import ReplacementGroup
from ..span import SpanGroup
from uttut.pipeline.edit import str2lst, lst2str


test_cases = [
    pytest.param(
        '我想要喝200元的珍奶10杯',
        ['我', '想要', '喝', '200元', '的', '珍奶', '10杯'],
        SpanGroup.add_all(
            [(0, 1), (1, 3), (3, 4), (4, 8), (8, 9), (9, 11), (11, 14)]),
        id='jieba',
    ),
    pytest.param(
        '我想要喝 200 元的珍奶 10 杯',
        ['我', '想要', '喝', ' ', '200', ' ', '元', '的', '珍奶', ' ', '10', ' ', '杯'],
        SpanGroup.add_all(
            [(0, 1), (1, 3), (3, 4), (4, 5), (5, 8), (8, 9), (9, 10),
             (10, 11), (11, 13), (13, 14), (14, 16), (16, 17), (17, 18)]),
        id='jieba with space',
    ),
    pytest.param(
        "I'vebeentoJapan.",
        ["I", "'", "ve", "been", "to", "Japan", "."],
        SpanGroup.add_all(
            [(0, 1), (1, 2), (2, 4), (4, 8), (8, 10), (10, 15), (15, 16)]),
        id='nltk punct after str2tr',
    ),
    pytest.param(
        "I'vebeentoJapan.",
        ["I've", "been", "to", "Japan", "."],
        SpanGroup.add_all(
            [(0, 4), (4, 8), (8, 10), (10, 15), (15, 16)]),
        id='nltk after str2tr',
    ),
]


@pytest.mark.parametrize("input_str,tokens,span_group,", test_cases)
def test_forward_apply(input_str, tokens, span_group):
    output = str2lst.apply(input_str, span_group)
    assert tokens == output


@pytest.mark.parametrize("input_str,tokens,span_group,", test_cases)
def test_backward_apply(input_str, tokens, span_group):
    output = lst2str.apply(tokens, span_group)
    assert input_str == output


def test_incompatible_str2lst():
    with pytest.raises(ValueError):
        str2lst.apply('薄餡亂入', SpanGroup.add_all([(0, 1), (1, 2)]))


@pytest.mark.parametrize(
    'input_lst,span_group',
    [
        pytest.param(['薄餡', '亂入'], SpanGroup.add_all([(0, 1)]), id='length'),
        pytest.param(['我', '想要', '喝', '多多綠'],
                     SpanGroup.add_all([(0, 1), (1, 3), (3, 5), (5, 8)]), id='element'),
    ],
)
def test_incompatible_lst2str(input_lst, span_group):
    with pytest.raises(ValueError):
        lst2str.apply(input_lst, span_group)


@pytest.mark.parametrize("input_str,tokens,span_group,", test_cases)
def test_gen_span_group(input_str, tokens, span_group):
    output = str2lst.gen_span_group(input_str, tokens)
    assert span_group == output


@pytest.mark.parametrize(
    'input_str,tokens',
    [
        pytest.param('我想要喝200元的珍奶10杯', ['薄餡', '亂入'], id='zh_all_mismatch'),
        pytest.param("I've been to Japan.", ["I", "'", "ve", "been", "to", "Japan", "."],
                     id='us lack of space'),
    ],
)
def test_gen_span_group_fail(input_str, tokens):
    with pytest.raises(ValueError, match='input_str and tokens are not compatible'):
        str2lst.gen_span_group(input_str, tokens)


@pytest.mark.parametrize(
    "input_str,tokens,replacement_group",
    [
        pytest.param(
            '我想要喝200元的珍奶10杯',
            ['我', '想要', '喝', '200元', '的', '珍奶', '10杯'],
            ReplacementGroup.add_all([]),
            id='jieba',
        ),
        pytest.param(
            '我想要喝 200 元的珍奶 10 杯',
            ['我', '想要', '喝', ' ', '200', ' ', '元', '的', '珍奶', ' ', '10', ' ', '杯'],
            ReplacementGroup.add_all([]),
            id='jieba with space',
        ),
        pytest.param(
            "I've been to Japan.",
            ["I", "'", "ve", "been", "to", "Japan", "."],
            ReplacementGroup.add_all([(4, 5, ''), (9, 10, ''), (12, 13, '')]),
            id='nltk punct True',
        ),
        pytest.param(
            "I've been to Japan.",
            ["I've", "been", "to", "Japan", "."],
            ReplacementGroup.add_all([(4, 5, ''), (9, 10, ''), (12, 13, '')]),
            id='nltk punct False',
        ),
    ],
)
def test_gen_replacement_group(input_str, tokens, replacement_group):
    output = str2lst.gen_replacement_group(input_str, tokens)
    assert replacement_group == output


@pytest.mark.parametrize(
    'input_str,tokens',
    [
        pytest.param('我想要喝200元的珍奶10杯', ['薄餡', '亂入'], id='zh_all_mismatch'),
        pytest.param('我想要喝珍奶', ['我', '想要', '喝', '多多綠'], id='zh_part_mismatch'),
        pytest.param('我想要喝珍奶', ['珍奶', '我', '想要', '喝'], id='zh_different_order'),
    ],
)
def test_gen_replacement_group_fail(input_str, tokens):
    with pytest.raises(ValueError, match='input_str and tokens are not compatible'):
        str2lst.gen_replacement_group(input_str, tokens)



from typing import Dict, List
import warnings

import json

from uttut.elements import Datum

from .intermediate import Intermediate
from .ops.base import LabelAligner, Operator
from .utils import unpack_datum


class Pipe:

    """Pipe is a container for a series of operators

    Attributes:
        _steps (Operators) : a list of operator instances
        _checkpoints (strs): a list of checkpoints
    """

    def __init__(self):
        self._steps = []
        self._checkpoints = {}

    def add(self, op_name: str, op_kwargs: Dict = None, checkpoint: str = None):
        """Add op into steps based on the operation name & kwargs.

        This method creates an operator given name & kwargs, and then append it
        to steps.

        Warning: This method will be deprecated, please use `add_op`.

        Args:
            op_name (str): the class name of operator.
            op_kwargs (dict): the keyword arguments to create operator.
            checkpoint (str): the name of checkpoint.

        Raises:
            TypeError: If input_type of op isn't consistent with pipe's output_type.
            KeyError: If the checkpoint name has already been added.
        """

        warnings.warn("This method will be deprecated, please use `add_op`.", DeprecationWarning)
        if op_kwargs is None:
            op_kwargs = {}

        op = Operator.from_dict({'op_name': op_name, 'op_kwargs': op_kwargs})
        self.add_op(op, checkpoint=checkpoint)

    def add_op(self, op: Operator, checkpoint: str = None):
        """Add op into steps

        Append the input op to steps.

        Args:
            op (Operator): An operator instance to be added to steps.
            checkpoint (str): the name of checkpoint

        Raises:
            TypeError: If input_type of operator isn't consistent with pipe's output_type.
            KeyError: If the checkpoint name has already been added.
        """

        if self.steps:
            if op.input_type != self.output_type:
                raise TypeError(
                    "InputType of the step op is not valid."
                    f"Got {op.input_type}, but requires {self.output_type}",
                )
        self._steps.append(op)

        if checkpoint is not None:
            if checkpoint in self.checkpoints:
                raise KeyError(f"duplicated checkpoints: {checkpoint}")
            self._checkpoints[checkpoint] = len(self.steps)

    def __eq__(self, other):
        return (self.steps, self._checkpoints) == (other.steps, other._checkpoints)

    @property
    def steps(self):
        return self._steps

    @property
    def checkpoints(self):
        return self._checkpoints

    @property
    def output_type(self):
        if not self.steps:
            raise IndexError("Pipe is empty!")
        return self.steps[-1].output_type

    def transform(self, datum: Datum):
        """Process data based on steps

        This method processes datum according to the Pipe's steps.

        Arg:
            datum (Datum)

        Returns:
            output_sequence: transfromed sequence
            intent_labels (ints)
            entity_labels (ints)
            label_alingers: an instance of LabelAlignerSequence
            intermediate: an instance of Intermediate

        """
        input_sequence, intent_labels, entity_labels = unpack_datum(datum)
        output_sequence, label_aligners, intermediate = self.transform_sequence(input_sequence)
        updated_entity_labels = label_aligners.transform(entity_labels)

        return output_sequence, intent_labels, updated_entity_labels, label_aligners, intermediate

    def transform_sequence(self, input_sequence):
        """Process input_sequence based on steps

        This method processes input_sequence according to the Pipe's steps.

        Arg:
            input_sequence

        Returns:
            output_sequence: transfromed sequence
            label_alingers: an instance of LabelAlignerSequence
            intermediate: an instance of Intermediate

        """
        intermediate = Intermediate(self.checkpoints)

        intermediate.add(input_sequence)
        label_aligners = LabelAlignerSequence()

        for op in self.steps:
            input_sequence, label_aligner = op.transform(input_sequence)
            intermediate.add(input_sequence)
            label_aligners.add(label_aligner)

        return input_sequence, label_aligners, intermediate

    def serialize(self) -> str:
        return json.dumps({
            'steps': [op.serialize() for op in self.steps],
            'checkpoints': self.checkpoints,
        })

    @classmethod
    def deserialize(cls, serialized_str: str) -> 'Pipe':
        pipe = cls()
        pipe_bundle = json.loads(serialized_str)
        # restore steps
        for step_info in pipe_bundle['steps']:
            op = Operator.deserialize(step_info)
            pipe.add_op(op)
        # restore checkpoints
        pipe._checkpoints = pipe_bundle['checkpoints']
        return pipe


class LabelAlignerSequence:

    def __init__(self):
        self.collections = []

    def add(self, label_aligner: LabelAligner):
        """Append label_aligner into collections

        Arg:
            label_aligner: an instance of LabelAligner

        """
        self.collections.append(label_aligner)

    def transform(self, labels: List[int]) -> List[int]:
        """Update labels based on given label_aligners

        Arg:
            labels (ints): raw labels (entity)

        Return:
            ints: updated labels which has the same length as
                  transformed sequence

        """
        for label_aligner in self.collections:
            labels = label_aligner.transform(labels)
        return labels

    def inverse_transform(self, labels: List[int]) -> List[int]:
        """Realign model prediction to its original input

        Arg:
            labels (ints): model prediction (entity)

        Return:
            ints: updated labels which has the same length as
                  original input (utterance)

        """
        for label_aligner in self.collections[::-1]:
            labels = label_aligner.inverse_transform(labels)
        return labels



import os
from setuptools import setup, find_packages, Extension
from pathlib import Path

LINE_TRACE = bool(os.environ.get('LINE_TRACE', 0))


try:
    long_description = open("README.md").read()
except IOError:
    long_description = ""

# a good way to structure python package with Cython
# https://stackoverflow.com/questions/4505747/how-should-i-structure-a-python-package-that-contains-cython-code
try:
    from Cython.Build import cythonize

except ImportError:
    use_cython = False
else:
    use_cython = True

ext_modules = []

if use_cython:
    ext_modules += [
        Extension('uttut.elements', ['uttut/elements.pyx']),
        Extension('uttut.expand_by_entities', ['uttut/expand_by_entities.pyx']),
        Extension('uttut.toolkits.get_kth_combination', ['uttut/toolkits/get_kth_combination.pyx']),
        Extension('uttut.toolkits.partition_by_entities', [
                  'uttut/toolkits/partition_by_entities.pyx']),
        Extension('uttut.pipeline.edit.replacement', ['uttut/pipeline/edit/replacement.pyx']),
        Extension('uttut.pipeline.edit.span', ['uttut/pipeline/edit/span.pyx']),
        Extension('uttut.pipeline.edit.utils', ['uttut/pipeline/edit/utils.pyx']),
        Extension('uttut.pipeline.edit.validation', ['uttut/pipeline/edit/validation.pyx']),
        Extension('uttut.pipeline.edit.label_propagation',
                  ['uttut/pipeline/edit/label_propagation.pyx']),
        Extension(
            name='uttut.pipeline.ops.utils.consistent_hash',
            sources=[
                'uttut/pipeline/ops/utils/consistent_hash.pyx',
                'uttut/pipeline/ops/utils/MurmurHash3.cpp',
            ],
            extra_compile_args=['-O3'],
            language='c++',
        ),
    ]
    ext_modules = cythonize(
        ext_modules,
        compiler_directives={
            'language_level': 3,
            'linetrace': LINE_TRACE,
            'profile': True,
            'binding': LINE_TRACE,
        },
    )

else:
    ext_modules += [
        Extension('uttut.elements', ['uttut/elements.c']),
        Extension('uttut.expand_by_entities', ['uttut/expand_by_entities.c']),
        Extension('uttut.toolkits.get_kth_combination', ['uttut/toolkits/get_kth_combination.c']),
        Extension('uttut.toolkits.partition_by_entities',
                  ['uttut/toolkits/partition_by_entities.c']),
        Extension('uttut.pipeline.edit.replacement', ['uttut/pipeline/edit/replacement.c']),
        Extension('uttut.pipeline.edit.span', ['uttut/pipeline/edit/span.c']),
        Extension('uttut.pipeline.edit.utils', ['uttut/pipeline/edit/utils.c']),
        Extension('uttut.pipeline.edit.validation', ['uttut/pipeline/edit/validation.c']),
        Extension('uttut.pipeline.edit.label_propagation',
                  ['uttut/pipeline/edit/label_propagation.c']),
        Extension('uttut.pipeline.ops.utils.consistent_hash',
                  ['uttut/pipeline/ops/utils/consistent_hash.cpp',
                   'uttut/pipeline/ops/utils/MurmurHash3.cpp']),
    ]


readme = Path(__file__).parent.joinpath('README.md')
if readme.exists():
    with readme.open() as f:
        long_description = f.read()
        try:
            from pypandoc import convert_text
            long_description = convert_text(
                long_description, 'rst', format='md')
        except ImportError:
            print("warning: pypandoc module not found, could not convert Markdown to RST")
else:
    long_description = '-'


setup(
    name="uttut",
    version='1.4.0',
    description="Yoctol Utterance processing utilities",
    license="MIT",
    author="cph",
    url='https://github.com/Yoctol/uttut',
    packages=find_packages(),
    install_requires=[
    ],
    long_description=long_description,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Development Status :: 3 - Alpha",
    ],
    ext_modules=ext_modules,
    python_requires='>=3.5',
)



from typing import Tuple

import re

from .base import Operator, NullLabelAligner
from ..edit.replacement import ReplacementGroup
from ..edit import str2str


class Lowercase(Operator):

    """
    Recognize uppercase characters and convert them into lowercase characters

    E.g.
    >>> from uttut.pipeline.ops.lowercase import Lowercase
    >>> op = Lowercase()
    >>> output_seq, label_aligner = op.transform("ABc")
    >>> output_labels = label_aligner.transform([1, 2, 3])
    >>> output_seq
    "abc"
    >>> output_labels
    [1, 2, 3]
    >>> label_aligner.inverse_transform(output_labels)
    [1, 2, 3]

    """

    _input_type = str
    _output_type = str

    REGEX_PATTERN = re.compile(r"[A-Z]+")

    def _transform(self, input_sequence: str) -> Tuple[str, 'LabelAligner']:
        forward_replacement_group = self._gen_forward_replacement_group(input_sequence)
        output_sequence = str2str.apply(input_sequence, forward_replacement_group)
        assert len(input_sequence) == len(output_sequence)

        label_aligner = LowercaseAligner(
            input_sequence=input_sequence,
            edit=forward_replacement_group,
            output_length=len(output_sequence),
        )
        return output_sequence, label_aligner

    def _gen_forward_replacement_group(self, input_str: str) -> ReplacementGroup:
        shift = 0
        replacement_group = ReplacementGroup()

        while shift < len(input_str):
            match = self.REGEX_PATTERN.search(input_str, shift)
            if match is None:
                break
            matched_str = input_str[match.start(): match.end()]
            replacement_group.add(
                start=match.start(),
                end=match.end(),
                new_value=matched_str.lower(),
                annotation='lowercase',
            )
            shift = match.end()
        replacement_group.done()
        return replacement_group


LowercaseAligner = NullLabelAligner



from typing import List

import re

from ..label_transducer import get_most_common, get_most_common_except_not_entity
from ..tokens import INT_TOKEN
from .base import PatternRecognizer, PatternRecognizerAligner


class IntTokenAligner(PatternRecognizerAligner):

    def _forward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_most_common(labels=labels, output_size=output_size)

    def _backward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_most_common_except_not_entity(labels, output_size)


class IntToken(PatternRecognizer):
    """
    Recognize integer (ex: 12, 10000) in the input string
    and replace them with INT_TOKEN (_int_)

    E.g.
    >>> from uttut.pipeline.ops.int_token import IntToken
    >>> op = IntToken()
    >>> output_seq, realigner = op.transform("10")
    >>> output_labels = label_aligner.transform([1, 1])
    >>> output_seq
    "_int_"
    >>> output_labels
    [1, 1, 1, 1, 1]
    >>> label_aligner.inverse_transform(output_labels)
    [1, 1]

    """

    _label_aligner_class = IntTokenAligner

    REGEX_PATTERN = re.compile(r"(?<![\.\d])\d+(?![\.\d])")
    TOKEN = INT_TOKEN

    def _gen_forward_replacement_group(self, input_str: str):  # type: ignore
        return super()._gen_forward_replacement_group(
            input_str=input_str,
            annotation='int-token',
        )



from .int_token import IntToken
from .float_token import FloatToken
from .int_token_with_space import IntTokenWithSpace
from .float_token_with_space import FloatTokenWithSpace
from .merge_whitespace_characters import MergeWhiteSpaceCharacters
from .strip_whitespace_characters import StripWhiteSpaceCharacters



import pytest

from ..replacement import Replacement


@pytest.mark.parametrize(
    "input_tuple",
    [
        pytest.param((10, 15, '薄餡亂入'), id='str-replacement'),
        pytest.param((10, 15, '薄餡亂入', 'ohoh'), id='str-replacement with annotation'),
        pytest.param((10, 15, ['薄餡亂入']), id='list-replacement'),
        pytest.param((10, 15, ['薄餡亂入'], 'ohoh'), id='list-replacement with annotation'),
    ],
)
def test_correctly_init(input_tuple):
    replacement = Replacement(*input_tuple)

    start = input_tuple[0]
    end = input_tuple[1]
    new_value = input_tuple[2]
    annotation = None
    if len(input_tuple) == 4:
        annotation = input_tuple[3]

    assert start == replacement.start
    assert end == replacement.end
    assert new_value == replacement.new_value
    assert annotation == replacement.annotation


@pytest.mark.parametrize(
    "obj1,obj2",
    [
        pytest.param((10, 15, '薄餡亂入'), (10, 15, '薄餡亂入'),
                     id='no annotation'),
        pytest.param((10, 15, '薄餡亂入', '1'), (10, 15, '薄餡亂入', '2'),
                     id='different annotation'),
    ],
)
def test_equal(obj1, obj2):
    assert Replacement(*obj1) == Replacement(*obj2)


@pytest.mark.parametrize(
    "obj1,obj2",
    [
        pytest.param((10, 15, '薄餡亂入'), (10, 15, '隼興亂入'), id='different new_value'),
        pytest.param((10, 15, '薄餡亂入'), (9, 15, '薄餡亂入'), id='different start index'),
        pytest.param((10, 15, '薄餡亂入'), (10, 11, '薄餡亂入'), id='different end index'),
        pytest.param((10, 15, ['薄餡亂入']), (10, 15, '薄餡亂入'), id='different type of new_value'),
    ],
)
def test_not_equal(obj1, obj2):
    assert Replacement(*obj1) != Replacement(*obj2)


@pytest.mark.parametrize(
    "input_tuple,expected",
    [
        pytest.param((10, 15, '薄餡亂入'), "(10, 15) => '薄餡亂入'", id='str-replacement'),
        pytest.param((10, 15, ['薄餡亂入']),
                     "(10, 15) => '['薄餡亂入']'", id='lst-replacement'),
        pytest.param((10, 15, 10000),
                     "(10, 15) => '10000'", id='int-replacement'),
    ],
)
def test_str(input_tuple, expected):
    replacement = Replacement(*input_tuple)
    assert expected == str(replacement)


@pytest.mark.parametrize(
    "input_tuple,expected",
    [
        pytest.param((10, 15, '薄餡亂入'),
                     "Replacement(10, 15, 薄餡亂入, annotation=None)",
                     id='str-replacement no annotation'),
        pytest.param((10, 15, '薄餡亂入', 'ohoh'),
                     "Replacement(10, 15, 薄餡亂入, annotation=ohoh)",
                     id='str-replacement with annotation'),
        pytest.param((10, 15, ['薄餡亂入']),
                     "Replacement(10, 15, ['薄餡亂入'], annotation=None)",
                     id='lst-replacement no annotation'),
        pytest.param((10, 15, ['薄餡亂入'], 'ohoh'),
                     "Replacement(10, 15, ['薄餡亂入'], annotation=ohoh)",
                     id='lst-replacement with annotation'),
        pytest.param((10, 15, 10000),
                     "Replacement(10, 15, 10000, annotation=None)",
                     id='int-replacement no annotation'),
        pytest.param((10, 15, 10000, 'ohoh'),
                     "Replacement(10, 15, 10000, annotation=ohoh)",
                     id='int-replacement with annotation'),
    ],
)
def test_representation(input_tuple, expected):
    replacement = Replacement(*input_tuple)
    assert expected == repr(replacement)



from uttut.elements import Datum
from typing import List, Tuple, Union


def partition_by_entities(
        datum: Datum,
        include_orig: bool,
    ) -> Tuple[List[List[str]], List[Union[None, str]]]:

    start = 0
    parts = []
    entity_labels = []
    utt = datum.utterance

    for entity in datum.entities:
        if start != entity.start:
            parts.append([utt[start: entity.start]])
            entity_labels.append(None)

        if entity.no_replacements():
            part_candidates = [entity.value]
        else:
            part_candidates = sorted(list(entity.replacements))
            if include_orig:
                part_candidates.append(entity.value)
        parts.append(part_candidates)
        entity_labels.append(entity.label)

        start = entity.end

    if start != len(utt):
        parts.append([utt[start:]])
        entity_labels.append(None)

    return parts, entity_labels



import pytest

from ...tests.common_tests import OperatorTestTemplate, ParamTuple
from ..int_token import IntToken


class TestIntToken(OperatorTestTemplate):

    params = [
        ParamTuple(
            "12 24 3666",
            [1, 1, 0, 2, 2, 0, 3, 3, 3, 3],
            "_int_ _int_ _int_",
            [1, 1, 1, 1, 1, 0, 2, 2, 2, 2, 2, 0, 3, 3, 3, 3, 3],
            id='int int int',
        ),
        ParamTuple(
            "12.3 1000 3.5",
            [1, 1, 1, 1, 0, 2, 2, 2, 2, 0, 3, 3, 3],
            "12.3 _int_ 3.5",
            [1, 1, 1, 1, 0, 2, 2, 2, 2, 2, 0, 3, 3, 3],
            id='float int float',
        ),
        ParamTuple(
            "999",
            [0, 0, 0],
            "_int_",
            [0, 0, 0, 0, 0],
            id='int with label 0',
        ),
        ParamTuple(
            "GB亂入",
            [1, 1, 2, 2],
            "GB亂入",
            [1, 1, 2, 2],
            id='identity',
        ),
        ParamTuple(
            "GB亂入10次",
            [1, 1, 2, 2, 3, 3, 4],
            "GB亂入_int_次",
            [1, 1, 2, 2, 3, 3, 3, 3, 3, 4],
            id='zh with int',
        ),
    ]

    @pytest.fixture(scope='class')
    def op(self):
        return IntToken()

    def test_equal(self, op):
        assert IntToken() == op



from typing import List

from ..ops.base import Operator, LabelAligner


class MockLabelAligner(LabelAligner):

    def _transform(self, labels):
        return labels

    def _inverse_transform(self, labels):
        return labels


class Str2Str(Operator):

    _input_type = str
    _output_type = str

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def _transform(self, input_sequence: str):  # type: ignore
        return input_sequence, MockLabelAligner(
            edit={},
            input_sequence=input_sequence,
            output_length=len(input_sequence),
        )


class Lst2Lst(Operator):

    _input_type = list
    _output_type = list

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def _transform(self, input_sequence: List[str]):  # type: ignore
        return input_sequence, MockLabelAligner(
            edit={},
            input_sequence=input_sequence,
            output_length=len(input_sequence),
        )


class Str2Lst(Operator):

    _input_type = str
    _output_type = list

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def _transform(self, input_sequence: str):  # type: ignore
        output_sequence = list(input_sequence)
        return output_sequence, MockLabelAligner(
            edit={},
            input_sequence=input_sequence,
            output_length=len(output_sequence),
        )



from typing import List
import warnings

from .validation cimport (  # noqa: E211, E999
    _validate_start_end_in_c,
    _validate_disjoint_in_c,
)


cdef class Replacement:

    '''Represents an edit to a Sequence such as a string or a list

    A sequence is a collection of objects that can be indexed in a linear fashion. E.g. a string

    A `Replacement` stores the data for substitution of values in a sequence.
    It represents the replacing a continuous slice of the sequence with the items of `new_value`.
    (The semantics are similar to Python's built-in slice assignment feature
     https://docs.python.org/3/reference/simple_stmts.html#assignment-statements ,
     but more restrictive in that negative integers are not allowed for `start` and `end`.)


    E.g.
    The transformation from "" to "abc": Replacement(0, 0, "abc")
    The transformation from "abc" to "aBc": Replacement(1, 2, "B")
    The transformation from "aBc" to "aBcdef": Replacement(3, 3, "def")
    The transformation from "aBcdef" to "": Replacement(0, 6, "")
    Note that insertion at position i is represented as Replacement(i, i, ...)
    ("Insertion at i" means that the original i-th element is moved to the right;
     similar to Python's list.insert)

    Args:
        start (int): start index (inclusive)
        end (int): end index (exclusive)
        new_value (obj): the new subsequence
        annotation (str, optional): an annotation for this replacement
    '''

    def __cinit__(
            self,
            unsigned int start,
            unsigned int end,
            object new_value,
            object annotation=None,
        ):
        _validate_start_end_in_c(start, end)
        self.start = start
        self.end = end
        self.new_value = new_value
        self.annotation = annotation

    def __eq__(self, other):
        same_start = other.start == self.start
        same_end = other.end == self.end
        same_new_values = other.new_value == self.new_value
        return same_start and same_end and same_new_values

    def __str__(self):
        cdef str output_str
        output_str = f"({self.start}, {self.end}) => '{self.new_value}'"
        return output_str

    def __repr__(self):
        cdef str representation
        representation = f"Replacement({self.start}, {self.end}, " \
            f"{self.new_value}, annotation={self.annotation})"
        return representation

    def __hash__(self):
        cdef int hash_num
        hash_num = hash((self.start, self.end))
        return hash_num


cdef class ReplacementGroup:   # noqa: E999

    def __cinit__(self):
        self._replacements = []
        self._is_done = False

    cpdef void add(  # type: ignore
            self,
            unsigned int start,
            unsigned int end,
            object new_value,
            object annotation=None,
        ):
        cdef Replacement replacement

        replacement = Replacement(
            start=start,
            end=end,
            new_value=new_value,
            annotation=annotation,
        )
        self._replacements.append(replacement)

    cpdef void done(self, bint skip_sort=False) except *:

        cdef unsigned int n_replacements

        self._replacements = list(set(self._replacements))

        n_replacements = len(self._replacements)

        if (n_replacements != 0) and (not skip_sort):
            self._validate_new_values()
            self._replacements = _sort_replacements(self._replacements)
            _validate_disjoint_in_c(self._replacements)

        self._is_done = True

    cdef void _validate_new_values(self) except *:
        target_type = type(self._replacements[0].new_value)
        for i, rep in enumerate(self._replacements):
            if not isinstance(rep.new_value, target_type):
                raise TypeError(
                    f"the new_value of {i}-th element is not a(an) {target_type.__name__}")

    @classmethod
    def add_all(cls, list replacements):  # type: ignore
        '''
        replacements = [
            (
                start: int,
                end: int,
                new_value: str or tokens,
                annotation: str, # optional
            ),
            ...
        ]
        '''
        cdef ReplacementGroup replacement_group
        cdef tuple replacement

        replacement_group = cls()
        for replacement in replacements:
            replacement_group.add(*replacement)
        replacement_group.done()
        return replacement_group

    cpdef bint is_empty(self):
        return len(self._replacements) == 0

    def __eq__(self, other):
        self._warn_not_done()
        if not isinstance(other, ReplacementGroup):
            return False
        same_length = len(other) == len(self._replacements)
        same_elements = set(other) == set(self._replacements)
        return same_length and same_elements

    def __getitem__(self, value):
        if not self._is_done:
            raise RuntimeError('Please call `done` first.')
        return self._replacements[value]

    def __len__(self):
        self._warn_not_done()
        return len(self._replacements)

    def _warn_not_done(self):
        if not self._is_done:
            warnings.warn('ReplacementGroup needs validation, please call `done`.')

    def __repr__(self):
        cdef int n_elements
        n_elements = len(self._replacements)
        return f"ReplacementGroup has {n_elements} elements"


cdef list _sort_replacements(list replacements):

    cdef Replacement e

    replacements = sorted(replacements, key=lambda e: e.end)
    replacements = sorted(replacements, key=lambda e: e.start)
    return replacements



import pytest

from uttut.elements import Datum, Intent, Entity
from ..pipe import Pipe


@pytest.fixture
def fake_pipe():
    p_custom = Pipe()
    p_custom.add('Str2Str', {'1': 1})
    p_custom.add('Str2Lst', {})
    p_custom.add('Lst2Lst', {})
    return p_custom


@pytest.fixture
def fake_pipe_with_checkpoints():
    p_custom = Pipe()
    p_custom.add('Str2Str', {'1': 1}, checkpoint='1')
    p_custom.add('Str2Lst', {})
    p_custom.add('Lst2Lst', {}, checkpoint='2')
    return p_custom


@pytest.fixture
def dummy_datum():
    return Datum(
        utterance='123',
        intents=[Intent(1)],
        entities=[
            Entity(label=1, start=0, end=1, value='1'),
            Entity(label=2, start=1, end=2, value='2'),
            Entity(label=3, start=2, end=3, value='3'),
        ],
    )


def test_pipe_has_invalid_op():
    p = Pipe()
    p.add('Lst2Lst')
    with pytest.raises(TypeError):
        p.add('Str2Str')


def test_pipe_has_duplicated_checkpoints():
    p = Pipe()
    p.add('Str2Str', checkpoint='1')
    with pytest.raises(KeyError):
        p.add('Str2Lst', checkpoint='1')


def test_pipe_can_have_duplicated_ops():
    p = Pipe()
    p.add('Str2Str', checkpoint='1')
    p.add('Str2Str', checkpoint='2')


def test_transform(fake_pipe, dummy_datum):
    output_seq, intent_labels, entity_labels, label_aligner, intermediate = fake_pipe.transform(
        dummy_datum)

    assert output_seq == ['1', '2', '3']
    assert intent_labels == [1]
    assert entity_labels == [1, 2, 3]
    assert ['123', '123', ['1', '2', '3'], ['1', '2', '3']] == intermediate[:]

    output = label_aligner.inverse_transform(entity_labels)
    assert output == [1, 2, 3]


def test_transform_with_checkpoints(fake_pipe_with_checkpoints, dummy_datum):
    output = fake_pipe_with_checkpoints.transform(dummy_datum)
    output_seq, intent_labels, entity_labels, label_aligner, intermediate = output

    assert output_seq == ['1', '2', '3']
    assert intent_labels == [1]
    assert entity_labels == [1, 2, 3]

    # intermediate
    assert '123' == intermediate.get_from_checkpoint('1')
    assert ['1', '2', '3'] == intermediate.get_from_checkpoint('2')
    assert ['123', '123', ['1', '2', '3'], ['1', '2', '3']] == intermediate[:]

    with pytest.raises(KeyError):
        intermediate.get_from_checkpoint('薄餡亂入')

    # realign labels
    output = label_aligner.inverse_transform(entity_labels)
    assert output == [1, 2, 3]


def test_serialization(fake_pipe):
    serialized_str = fake_pipe.serialize()
    o_pipe = Pipe.deserialize(serialized_str)
    assert fake_pipe == o_pipe


def test_serialization_with_checkpoint(fake_pipe_with_checkpoints):
    serialized_str = fake_pipe_with_checkpoints.serialize()
    o_pipe = Pipe.deserialize(serialized_str)
    assert fake_pipe_with_checkpoints == o_pipe



import pytest

from .tokenization import BasicTokenizer as BertBasicTokenizer
from ..basic import basic_pipe

from uttut.elements import Datum


test_cases = [
    pytest.param(
        "我想要喝珍奶",
        id='zh',
    ),
    pytest.param(  # ref: Bert tokenization_test.py#L62
        " \tHeLLo!how  \n Are yoU?  ",
        id='eng',
    ),
    pytest.param(  # ref: Bert tokenization_test.py#L64
        u"H\u00E9llo",
        id='eng + unicode',
    ),
    pytest.param(
        u"ah\u535A\u63A8zz",
        id="eng + zh",  # ref: Bert tokenization_test.py#L55
    ),
]


@pytest.fixture
def tokenizer():
    yield BertBasicTokenizer()


@pytest.mark.parametrize("input_str", test_cases)
def test_all(input_str, tokenizer):
    datum = Datum(input_str)
    output, _, _, _, _ = basic_pipe.transform(datum)
    expected_output = tokenizer.tokenize(input_str)
    assert expected_output == output



from typing import List

import re

from ..label_transducer import get_most_common
from .base import PatternRecognizer, PatternRecognizerAligner


class MergeWhiteSpaceCharactersAligner(PatternRecognizerAligner):

    def _forward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_most_common(labels=labels, output_size=output_size)

    def _backward_transduce_func(self, labels: List[int], output_size: int) -> List[int]:
        return get_most_common(labels=labels, output_size=output_size)


class MergeWhiteSpaceCharacters(PatternRecognizer):

    """
    Recognize contiguous whitespace characters in the string
    and replace it with a whitespace character (" ")

    E.g.
    >>> from uttut.pipeline.ops.merge_whitespace_characters import MergeWhiteSpaceCharacters
    >>> op = MergeWhiteSpaceCharacters()
    >>> output_seq, label_aligner = op.transform("\n\n  \t\t")
    >>> output_labels = label_aligner.transform([1, 1, 1, 1, 1, 1])
    >>> output_seq
    " "
    >>> output_labels
    [0]
    >>> label_aligner(output_labels)
    [0, 0, 0, 0, 0, 0]

    """

    _label_aligner_class = MergeWhiteSpaceCharactersAligner

    REGEX_PATTERN = re.compile(r"\s+")
    TOKEN = " "



from unittest import TestCase
from functools import reduce
from operator import mul

from ..get_kth_combination import get_kth_combination


class GetKthCombinationTestCase(TestCase):

    def test_get_kth_combination(self):
        iterables = [[1, 2, 3], ['A', 'B', 'C'], ['I', 'II']]
        expected_outputs = [
            [1, 'A', 'I'],
            [2, 'A', 'I'],
            [3, 'A', 'I'],
            [1, 'B', 'I'],
            [2, 'B', 'I'],
            [3, 'B', 'I'],
            [1, 'C', 'I'],
            [2, 'C', 'I'],
            [3, 'C', 'I'],
            [1, 'A', 'II'],
            [2, 'A', 'II'],
            [3, 'A', 'II'],
            [1, 'B', 'II'],
            [2, 'B', 'II'],
            [3, 'B', 'II'],
            [1, 'C', 'II'],
            [2, 'C', 'II'],
            [3, 'C', 'II'],
        ]

        for i, expected_output in enumerate(expected_outputs):
            with self.subTest(k=i):
                output = get_kth_combination(iterables, i)
                self.assertEqual(expected_output, output)

        # no duplicated output when k <= n_combinations
        n_combinations = reduce(mul, [len(e) for e in iterables])
        self.assertEqual(
            len(
                set(
                    [
                        str(get_kth_combination(iterables, k))
                        for k in range(n_combinations)
                    ],
                ),
            ),
            n_combinations,
        )



from unittest import TestCase

from uttut.elements import Datum, Entity, Intent
from ..partition_by_entities import partition_by_entities


class PartitionByEntitiesTestCase(TestCase):

    def setUp(self):
        self.utterance = '我想訂明天從紐約飛到新加坡的機票'
        self.entities = [
            Entity(label=0, value='明天', start=3, end=5, replacements=['下禮拜二']),
            Entity(label=1, value='紐約', start=6, end=8),
            Entity(label=2, value='新加坡', start=10, end=13, replacements=['斯堪地那維亞', 'KIX']),
        ]
        self.intents = [
            Intent(label=0),
        ]
        self.datum = Datum(
            utterance=self.utterance,
            intents=self.intents,
            entities=self.entities,
        )

        self.datum_wo_entity = Datum(
            utterance='薄餡亂入',
            intents=[Intent(label=0)],
        )

    def test_partition_by_entities(self):
        actual_parts, entity_names = partition_by_entities(self.datum, False)
        expected_parts = [
            ['我想訂'],
            ['下禮拜二'],
            ['從'],
            ['紐約'],
            ['飛到'],
            ['斯堪地那維亞', 'KIX'],
            ['的機票'],
        ]
        for exp_part, act_part in zip(expected_parts, actual_parts):
            self.assertEqual(set(exp_part), set(act_part))
        self.assertEqual(
            entity_names,
            [None, 0, None, 1, None, 2, None],
        )

    def test_partition_by_entities_include_orig(self):
        actual_parts, entity_names = partition_by_entities(self.datum, True)
        expected_parts = [
            ['我想訂'],
            ['明天', '下禮拜二'],
            ['從'],
            ['紐約'],
            ['飛到'],
            ['新加坡', '斯堪地那維亞', 'KIX'],
            ['的機票'],
        ]
        for exp_part, act_part in zip(expected_parts, actual_parts):
            self.assertEqual(set(exp_part), set(act_part))
        self.assertEqual(
            entity_names,
            [None, 0, None, 1, None, 2, None],
        )

    def test_datum_wo_entity(self):
        # do not include origin
        output = partition_by_entities(self.datum_wo_entity, True)
        self.assertEqual(([['薄餡亂入']], [None]), output)

        # include origin
        output = partition_by_entities(self.datum_wo_entity, False)
        self.assertEqual(([['薄餡亂入']], [None]), output)



import pytest

from ..span import Span, SpanGroup


@pytest.mark.filterwarnings("ignore")
def test_correctly_init():
    span_group = SpanGroup()
    assert span_group.is_empty() is True
    assert len(span_group) == 0


@pytest.mark.parametrize(
    "spans",
    [
        pytest.param([(0, 1), (3, 5)], id='not contiguous'),
        pytest.param([(1, 3), (3, 5)], id='not start from 0'),
    ],
)
def test_span_group_should_be_contiguous(spans):
    with pytest.raises(ValueError):
        SpanGroup.add_all(spans)


def test_add_all():
    objs = [(0, 0), (0, 2)]
    span_group = SpanGroup.add_all(objs)
    assert len(span_group) == len(objs)
    assert [Span(*obj) for obj in objs] == span_group[:]


def test_add_all_empty():
    span_group = SpanGroup.add_all([])
    assert span_group.is_empty() is True


def test_not_done_warning():
    span_group = SpanGroup()
    with pytest.warns(UserWarning, match='SpanGroup needs validation, please call `done`.'):
        len(span_group)


def test_not_done_error():
    span_group = SpanGroup()
    span_group.add(1, 3)
    with pytest.raises(RuntimeError, match='Please call `done` first.'):
        span_group[0]


@pytest.mark.parametrize(
    "obj,error_type",
    [
        pytest.param([(0,)], ValueError, id='lack of elements'),
        pytest.param([(0, 0, 3)], ValueError, id='has redundant elements'),
    ],
)
def test_add_fails(obj, error_type):
    with pytest.raises(error_type):
        SpanGroup.add_all(obj)


@pytest.mark.parametrize(
    "group1,group2",
    [
        pytest.param(SpanGroup.add_all([]), SpanGroup.add_all([]), id='empty'),
        pytest.param(SpanGroup.add_all([(0, 0), (0, 2)]),
                     SpanGroup.add_all([(0, 0), (0, 2)]), id='span'),
    ],
)
def test_equal(group1, group2):
    assert group1 == group2


@pytest.mark.parametrize(
    "group1,group2",
    [
        pytest.param(SpanGroup.add_all([(0, 0)]), [(0, 0)], id='different type'),
        pytest.param(SpanGroup.add_all([(0, 0)]), SpanGroup.add_all([(0, 0), (0, 2)]),
                     id='different length'),
        pytest.param(SpanGroup.add_all([(0, 0), (0, 2)]), SpanGroup.add_all([(0, 0), (0, 3)]),
                     id='different element'),
    ],
)
def test_not_equal(group1, group2):
    assert group1 != group2


@pytest.mark.parametrize(
    "objs,expected_objs",
    [
        pytest.param([(1, 2), (0, 1)], [Span(0, 1), Span(1, 2)]),
    ],
)
def test_need_sorted(objs, expected_objs):
    obj_group = SpanGroup.add_all(objs)
    assert expected_objs == obj_group[:]


@pytest.mark.parametrize(
    "objs",
    [
        pytest.param([(1, 10), (2, 15)], id='intersect'),
        pytest.param([(1, 10), (2, 8)], id='include'),
    ],
)
def test_validate_disjoint(objs):
    with pytest.raises(ValueError):
        SpanGroup.add_all(objs)


@pytest.mark.parametrize(
    "name,group",
    [
        pytest.param('SpanGroup', SpanGroup.add_all([(0, 0)]), id='str-span'),
    ],
)
def test_representation(name, group):
    assert name == repr(group)
